@@ -5,19 +5,19 @@
  *
  * ifenslave.c: Configure network interfaces for parallel routing.
  *
- *	This program controls the Linux implementation of running multiple
- *	network interfaces in parallel.
+ *      This program controls the Linux implementation of running multiple
+ *      network interfaces in parallel.
  *
- * Author:	Donald Becker <becker@cesdis.gsfc.nasa.gov>
- *		Copyright 1994-1996 Donald Becker
+ * Author:      Donald Becker <becker@cesdis.gsfc.nasa.gov>
+ *              Copyright 1994-1996 Donald Becker
  *
- *		This program is free software; you can redistribute it
- *		and/or modify it under the terms of the GNU General Public
- *		License as published by the Free Software Foundation.
+ *              This program is free software; you can redistribute it
+ *              and/or modify it under the terms of the GNU General Public
+ *              License as published by the Free Software Foundation.
  *
- *	The author may be reached as becker@CESDIS.gsfc.nasa.gov, or C/O
- *	Center of Excellence in Space Data and Information Sciences
- *	   Code 930.5, Goddard Space Flight Center, Greenbelt MD 20771
+ *      The author may be reached as becker@CESDIS.gsfc.nasa.gov, or C/O
+ *      Center of Excellence in Space Data and Information Sciences
+ *         Code 930.5, Goddard Space Flight Center, Greenbelt MD 20771
  *
  *  Changes :
  *    - 2000/10/02 Willy Tarreau <willy at meta-x.org> :
@@ -79,35 +79,40 @@
  *         current version is 1.0.0 as a base line.
  *
  *    - 2003/05/22 - Jay Vosburgh <fubar at us dot ibm dot com>
- *	 - ifenslave -c was broken; it's now fixed
- *	 - Fixed problem with routes vanishing from master during enslave
- *	   processing.
+ *       - ifenslave -c was broken; it's now fixed
+ *       - Fixed problem with routes vanishing from master during enslave
+ *         processing.
  *
  *    - 2003/05/27 - Amir Noam <amir.noam at intel dot com>
- *	 - Fix backward compatibility issues:
- *	   For drivers not using ABI versions, slave was set down while
- *	   it should be left up before enslaving.
- *	   Also, master was not set down and the default set_mac_address()
- *	   would fail and generate an error message in the system log.
- * 	 - For opt_c: slave should not be set to the master's setting
- *	   while it is running. It was already set during enslave. To
- *	   simplify things, it is now handeled separately.
+ *       - Fix backward compatibility issues:
+ *         For drivers not using ABI versions, slave was set down while
+ *         it should be left up before enslaving.
+ *         Also, master was not set down and the default set_mac_address()
+ *         would fail and generate an error message in the system log.
+ *       - For opt_c: slave should not be set to the master's setting
+ *         while it is running. It was already set during enslave. To
+ *         simplify things, it is now handeled separately.
  *
  *    - 2003/12/01 - Shmulik Hen <shmulik.hen at intel dot com>
- *	 - Code cleanup and style changes
- *	   set version to 1.1.0
+ *       - Code cleanup and style changes
+ *         set version to 1.1.0
  */
 
 #include "libbb.h"
 
+/* #include <net/if.h> - no. linux/if_bonding.h pulls in linux/if.h */
 #include <net/if_arp.h>
 #include <linux/if_bonding.h>
 #include <linux/sockios.h>
 
-typedef unsigned long long u64; /* hack, so we may include kernel's ethtool.h */
-typedef uint32_t u32;           /* ditto */
-typedef uint16_t u16;           /* ditto */
-typedef uint8_t u8;             /* ditto */
+#ifndef IFNAMSIZ
+#define IFNAMSIZ 16
+#endif
+
+typedef uint64_t u64; /* hack, so we may include kernel's ethtool.h */
+typedef uint32_t u32; /* ditto */
+typedef uint16_t u16; /* ditto */
+typedef uint8_t u8;   /* ditto */
 #include <linux/ethtool.h>
 
 
@@ -116,7 +121,7 @@ struct dev_data {
 };
 
 
-enum { skfd = 3 };      /* AF_INET socket for ioctl() calls.*/
+enum { skfd = 3 };      /* AF_INET socket for ioctl() calls. */
 struct globals {
 	unsigned abi_ver;       /* userland - kernel ABI version */
 	smallint hwaddr_set;    /* Master's hwaddr is set */
@@ -129,197 +134,131 @@ struct globals {
 #define master     (G.master    )
 #define slave      (G.slave     )
 #define INIT_G() do { \
-        SET_PTR_TO_GLOBALS(xzalloc(sizeof(G))); \
+	SET_PTR_TO_GLOBALS(xzalloc(sizeof(G))); \
 } while (0)
 
 
-static void get_drv_info(char *master_ifname);
-static int get_if_settings(char *ifname, struct dev_data *dd);
-static int get_slave_flags(char *slave_ifname);
-static int set_master_hwaddr(char *master_ifname, struct sockaddr *hwaddr);
-static int set_slave_hwaddr(char *slave_ifname, struct sockaddr *hwaddr);
-static int set_slave_mtu(char *slave_ifname, int mtu);
-static int set_if_flags(char *ifname, short flags);
-static int set_if_up(char *ifname, short flags);
-static int set_if_down(char *ifname, short flags);
-static int clear_if_addr(char *ifname);
-static int set_if_addr(char *master_ifname, char *slave_ifname);
-static void change_active(char *master_ifname, char *slave_ifname);
-static int enslave(char *master_ifname, char *slave_ifname);
-static int release(char *master_ifname, char *slave_ifname);
+/* NOINLINEs are placed where it results in smaller code (gcc 4.3.1) */
 
+static void strncpy_IFNAMSIZ(char *dst, const char *src)
+{
+	strncpy(dst, src, IFNAMSIZ);
+}
 
-int ifenslave_main(int argc, char **argv) MAIN_EXTERNALLY_VISIBLE;
-int ifenslave_main(int argc ATTRIBUTE_UNUSED, char **argv)
+static int ioctl_on_skfd(unsigned request, struct ifreq *ifr)
+{
+	return ioctl(skfd, request, ifr);
+}
+
+static int set_ifrname_and_do_ioctl(unsigned request, struct ifreq *ifr, const char *ifname)
+{
+	strncpy_IFNAMSIZ(ifr->ifr_name, ifname);
+	return ioctl_on_skfd(request, ifr);
+}
+
+static int get_if_settings(char *ifname, struct dev_data *dd)
 {
-	char *master_ifname, *slave_ifname;
-	int rv;
 	int res;
-	unsigned opt;
-	enum {
-		OPT_c = (1 << 0),
-		OPT_d = (1 << 1),
-		OPT_f = (1 << 2),
-	};
-#if ENABLE_GETOPT_LONG
-	static const char ifenslave_longopts[] ALIGN1 =
-		"change-active" No_argument "c"
-		"detach"        No_argument "d"
-		"force"         No_argument "f"
-	;
 
-	applet_long_options = ifenslave_longopts;
-#endif
-	opt = getopt32(argv, "cdf");
-	argv += optind;
-	if (opt & (opt-1)) /* options check */
-		bb_show_usage();
-
-	/* Copy the interface name */
-	master_ifname = *argv++;
-
-	/* No remaining args means show all interfaces. */
-	if (!master_ifname) {
-		display_interfaces(NULL);
-		return EXIT_SUCCESS;
-	}
-
-	/* Open a basic socket */
-	xmove_fd(xsocket(AF_INET, SOCK_DGRAM, 0), skfd);
-
-	/* exchange abi version with bonding module */
-	get_drv_info(master_ifname);
-
-	slave_ifname = *argv++;
-	if (!slave_ifname) {
-		if (opt & (OPT_d|OPT_c)) {
-			display_interfaces(slave_ifname);
-			return 2; /* why? */
-		}
-		/* A single arg means show the
-		 * configuration for this interface
-		 */
-		display_interfaces(master_ifname);
-		return EXIT_SUCCESS;
-	}
-
-	res = get_if_settings(master_ifname, &master);
-	if (res) {
-		/* Probably a good reason not to go on */
-		bb_perror_msg_and_die("%s: can't get settings", master_ifname);
-	}
-
-	/* check if master is indeed a master;
-	 * if not then fail any operation
-	 */
-	if (!(master.flags.ifr_flags & IFF_MASTER))
-		bb_error_msg_and_die("%s is not a master", master_ifname);
-
-	/* check if master is up; if not then fail any operation */
-	if (!(master.flags.ifr_flags & IFF_UP))
-		bb_error_msg_and_die("%s is not up", master_ifname);
-
-	/* Only for enslaving */
-	if (!(opt & (OPT_c|OPT_d))) {
-		sa_family_t master_family = master.hwaddr.ifr_hwaddr.sa_family;
-
-		/* The family '1' is ARPHRD_ETHER for ethernet. */
-		if (master_family != 1 && !(opt & OPT_f)) {
-			bb_error_msg_and_die(
-				"%s is not ethernet-like (-f overrides)",
-				master_ifname);
-		}
-	}
-
-	/* Accepts only one slave */
-	if (opt & OPT_c) {
-		/* change active slave */
-		res = get_slave_flags(slave_ifname);
-		if (res) {
-			bb_perror_msg_and_die(
-				"%s: can't get flags", slave_ifname);
-		}
-		change_active(master_ifname, slave_ifname);
-		return EXIT_SUCCESS;
-	}
-
-	/* Accept multiple slaves */
-	res = 0;
-	do {
-		if (opt & OPT_d) {
-			/* detach a slave interface from the master */
-			rv = get_slave_flags(slave_ifname);
-			if (rv) {
-				/* Can't work with this slave. */
-				/* remember the error and skip it*/
-				bb_perror_msg(
-					"skipping %s: can't get flags",
-					slave_ifname);
-				res = rv;
-				continue;
-			}
-			rv = release(master_ifname, slave_ifname);
-			if (rv) {
-				bb_perror_msg(
-					"master %s, slave %s: "
-					"can't release",
-					master_ifname, slave_ifname);
-				res = rv;
-			}
-		} else {
-			/* attach a slave interface to the master */
-			rv = get_if_settings(slave_ifname, &slave);
-			if (rv) {
-				/* Can't work with this slave. */
-				/* remember the error and skip it*/
-				bb_perror_msg(
-					"skipping %s: can't get settings",
-					slave_ifname);
-				res = rv;
-				continue;
-			}
-			rv = enslave(master_ifname, slave_ifname);
-			if (rv) {
-				bb_perror_msg(
-					"master %s, slave %s: "
-					"can't enslave",
-					master_ifname, slave_ifname);
-				res = rv;
-			}
-		}
-	} while ((slave_ifname = *argv++) != NULL);
-
-	if (ENABLE_FEATURE_CLEAN_UP) {
-		close(skfd);
-	}
+	res = set_ifrname_and_do_ioctl(SIOCGIFMTU, &dd->mtu, ifname);
+	res |= set_ifrname_and_do_ioctl(SIOCGIFFLAGS, &dd->flags, ifname);
+	res |= set_ifrname_and_do_ioctl(SIOCGIFHWADDR, &dd->hwaddr, ifname);
 
 	return res;
 }
 
-static void get_drv_info(char *master_ifname)
+static int get_slave_flags(char *slave_ifname)
+{
+	return set_ifrname_and_do_ioctl(SIOCGIFFLAGS, &slave.flags, slave_ifname);
+}
+
+static int set_hwaddr(char *ifname, struct sockaddr *hwaddr)
 {
 	struct ifreq ifr;
-	struct ethtool_drvinfo info;
 
-	memset(&ifr, 0, sizeof(ifr));
-	strncpy(ifr.ifr_name, master_ifname, IFNAMSIZ);
-	ifr.ifr_data = (caddr_t)&info;
+	memcpy(&(ifr.ifr_hwaddr), hwaddr, sizeof(*hwaddr));
+	return set_ifrname_and_do_ioctl(SIOCSIFHWADDR, &ifr, ifname);
+}
 
-	info.cmd = ETHTOOL_GDRVINFO;
-	strncpy(info.driver, "ifenslave", 32);
-	snprintf(info.fw_version, 32, "%d", BOND_ABI_VERSION);
+static int set_mtu(char *ifname, int mtu)
+{
+	struct ifreq ifr;
 
-	if (ioctl(skfd, SIOCETHTOOL, &ifr) < 0) {
-		if (errno == EOPNOTSUPP)
-			return;
-		bb_perror_msg_and_die("%s: SIOCETHTOOL error", master_ifname);
+	ifr.ifr_mtu = mtu;
+	return set_ifrname_and_do_ioctl(SIOCSIFMTU, &ifr, ifname);
+}
+
+static int set_if_flags(char *ifname, int flags)
+{
+	struct ifreq ifr;
+
+	ifr.ifr_flags = flags;
+	return set_ifrname_and_do_ioctl(SIOCSIFFLAGS, &ifr, ifname);
+}
+
+static int set_if_up(char *ifname, int flags)
+{
+	int res = set_if_flags(ifname, flags | IFF_UP);
+	if (res)
+		bb_perror_msg("%s: can't up", ifname);
+	return res;
+}
+
+static int set_if_down(char *ifname, int flags)
+{
+	int res = set_if_flags(ifname, flags & ~IFF_UP);
+	if (res)
+		bb_perror_msg("%s: can't down", ifname);
+	return res;
+}
+
+static int clear_if_addr(char *ifname)
+{
+	struct ifreq ifr;
+
+	ifr.ifr_addr.sa_family = AF_INET;
+	memset(ifr.ifr_addr.sa_data, 0, sizeof(ifr.ifr_addr.sa_data));
+	return set_ifrname_and_do_ioctl(SIOCSIFADDR, &ifr, ifname);
+}
+
+static int set_if_addr(char *master_ifname, char *slave_ifname)
+{
+#if (SIOCGIFADDR | SIOCSIFADDR \
+  | SIOCGIFDSTADDR | SIOCSIFDSTADDR \
+  | SIOCGIFBRDADDR | SIOCSIFBRDADDR \
+  | SIOCGIFNETMASK | SIOCSIFNETMASK) <= 0xffff
+#define INT uint16_t
+#else
+#define INT int
+#endif
+	static const struct {
+		INT g_ioctl;
+		INT s_ioctl;
+	} ifra[] = {
+		{ SIOCGIFADDR,    SIOCSIFADDR    },
+		{ SIOCGIFDSTADDR, SIOCSIFDSTADDR },
+		{ SIOCGIFBRDADDR, SIOCSIFBRDADDR },
+		{ SIOCGIFNETMASK, SIOCSIFNETMASK },
+	};
+
+	struct ifreq ifr;
+	int res;
+	unsigned i;
+
+	for (i = 0; i < ARRAY_SIZE(ifra); i++) {
+		res = set_ifrname_and_do_ioctl(ifra[i].g_ioctl, &ifr, master_ifname);
+		if (res < 0) {
+			ifr.ifr_addr.sa_family = AF_INET;
+			memset(ifr.ifr_addr.sa_data, 0,
+			       sizeof(ifr.ifr_addr.sa_data));
+		}
+
+		res = set_ifrname_and_do_ioctl(ifra[i].s_ioctl, &ifr, slave_ifname);
+		if (res < 0)
+			return res;
 	}
 
-	abi_ver = bb_strtou(info.fw_version, NULL, 0);
-	if (errno)
-		bb_error_msg_and_die("%s: SIOCETHTOOL error", master_ifname);
-
-	return;
+	return 0;
 }
 
 static void change_active(char *master_ifname, char *slave_ifname)
@@ -327,15 +266,12 @@ static void change_active(char *master_ifname, char *slave_ifname)
 	struct ifreq ifr;
 
 	if (!(slave.flags.ifr_flags & IFF_SLAVE)) {
-		bb_error_msg_and_die(
-			"%s is not a slave",
-			slave_ifname);
+		bb_error_msg_and_die("%s is not a slave", slave_ifname);
 	}
 
-	strncpy(ifr.ifr_name, master_ifname, IFNAMSIZ);
-	strncpy(ifr.ifr_slave, slave_ifname, IFNAMSIZ);
-	if (ioctl(skfd, SIOCBONDCHANGEACTIVE, &ifr) < 0
-	 && ioctl(skfd, BOND_CHANGE_ACTIVE_OLD, &ifr) < 0
+	strncpy_IFNAMSIZ(ifr.ifr_slave, slave_ifname);
+	if (set_ifrname_and_do_ioctl(SIOCBONDCHANGEACTIVE, &ifr, master_ifname)
+	 && ioctl_on_skfd(BOND_CHANGE_ACTIVE_OLD, &ifr)
 	) {
 		bb_perror_msg_and_die(
 			"master %s, slave %s: can't "
@@ -344,7 +280,7 @@ static void change_active(char *master_ifname, char *slave_ifname)
 	}
 }
 
-static int enslave(char *master_ifname, char *slave_ifname)
+static NOINLINE int enslave(char *master_ifname, char *slave_ifname)
 {
 	struct ifreq ifr;
 	int res;
@@ -378,7 +314,7 @@ static int enslave(char *master_ifname, char *slave_ifname)
 	}
 
 	if (master.mtu.ifr_mtu != slave.mtu.ifr_mtu) {
-		res = set_slave_mtu(slave_ifname, master.mtu.ifr_mtu);
+		res = set_mtu(slave_ifname, master.mtu.ifr_mtu);
 		if (res) {
 			bb_perror_msg("%s: can't set MTU", slave_ifname);
 			return res;
@@ -394,9 +330,7 @@ static int enslave(char *master_ifname, char *slave_ifname)
 			 * the application sets the slave's
 			 * hwaddr
 			 */
-			res = set_slave_hwaddr(slave_ifname,
-					       &(master.hwaddr.ifr_hwaddr));
-			if (res) {
+			if (set_hwaddr(slave_ifname, &(master.hwaddr.ifr_hwaddr))) {
 				bb_perror_msg("%s: can't set hw address",
 						slave_ifname);
 				goto undo_mtu;
@@ -405,8 +339,7 @@ static int enslave(char *master_ifname, char *slave_ifname)
 			/* For old ABI the application needs to bring the
 			 * slave back up
 			 */
-			res = set_if_up(slave_ifname, slave.flags.ifr_flags);
-			if (res)
+			if (set_if_up(slave_ifname, slave.flags.ifr_flags))
 				goto undo_slave_mac;
 		}
 		/* The driver is using a new ABI,
@@ -422,14 +355,11 @@ static int enslave(char *master_ifname, char *slave_ifname)
 			/* For old ABI, the master needs to be
 			 * down before setting it's hwaddr
 			 */
-			res = set_if_down(master_ifname, master.flags.ifr_flags);
-			if (res)
+			if (set_if_down(master_ifname, master.flags.ifr_flags))
 				goto undo_mtu;
 		}
 
-		res = set_master_hwaddr(master_ifname,
-					&(slave.hwaddr.ifr_hwaddr));
-		if (res) {
+		if (set_hwaddr(master_ifname, &(slave.hwaddr.ifr_hwaddr))) {
 			bb_error_msg("%s: can't set hw address",
 				master_ifname);
 			goto undo_mtu;
@@ -439,8 +369,7 @@ static int enslave(char *master_ifname, char *slave_ifname)
 			/* For old ABI, bring the master
 			 * back up
 			 */
-			res = set_if_up(master_ifname, master.flags.ifr_flags);
-			if (res)
+			if (set_if_up(master_ifname, master.flags.ifr_flags))
 				goto undo_master_mac;
 		}
 
@@ -448,29 +377,26 @@ static int enslave(char *master_ifname, char *slave_ifname)
 	}
 
 	/* Do the real thing */
-	strncpy(ifr.ifr_name, master_ifname, IFNAMSIZ);
-	strncpy(ifr.ifr_slave, slave_ifname, IFNAMSIZ);
-	if (ioctl(skfd, SIOCBONDENSLAVE, &ifr) < 0
-	 && ioctl(skfd, BOND_ENSLAVE_OLD, &ifr) < 0
+	strncpy_IFNAMSIZ(ifr.ifr_slave, slave_ifname);
+	if (set_ifrname_and_do_ioctl(SIOCBONDENSLAVE, &ifr, master_ifname)
+	 && ioctl_on_skfd(BOND_ENSLAVE_OLD, &ifr)
 	) {
-		res = 1;
-	}
-
-	if (res)
 		goto undo_master_mac;
+	}
 
 	return 0;
 
 /* rollback (best effort) */
  undo_master_mac:
-	set_master_hwaddr(master_ifname, &(master.hwaddr.ifr_hwaddr));
+	set_hwaddr(master_ifname, &(master.hwaddr.ifr_hwaddr));
 	hwaddr_set = 0;
 	goto undo_mtu;
+
  undo_slave_mac:
-	set_slave_hwaddr(slave_ifname, &(slave.hwaddr.ifr_hwaddr));
+	set_hwaddr(slave_ifname, &(slave.hwaddr.ifr_hwaddr));
  undo_mtu:
-	set_slave_mtu(slave_ifname, slave.mtu.ifr_mtu);
-	return res;
+	set_mtu(slave_ifname, slave.mtu.ifr_mtu);
+	return 1;
 }
 
 static int release(char *master_ifname, char *slave_ifname)
@@ -479,15 +405,13 @@ static int release(char *master_ifname, char *slave_ifname)
 	int res = 0;
 
 	if (!(slave.flags.ifr_flags & IFF_SLAVE)) {
-		bb_error_msg("%s is not a slave",
-			slave_ifname);
+		bb_error_msg("%s is not a slave", slave_ifname);
 		return 1;
 	}
 
-	strncpy(ifr.ifr_name, master_ifname, IFNAMSIZ);
-	strncpy(ifr.ifr_slave, slave_ifname, IFNAMSIZ);
-	if (ioctl(skfd, SIOCBONDRELEASE, &ifr) < 0
-	 && ioctl(skfd, BOND_RELEASE_OLD, &ifr) < 0
+	strncpy_IFNAMSIZ(ifr.ifr_slave, slave_ifname);
+	if (set_ifrname_and_do_ioctl(SIOCBONDRELEASE, &ifr, master_ifname) < 0
+	 && ioctl_on_skfd(BOND_RELEASE_OLD, &ifr) < 0
 	) {
 		return 1;
 	}
@@ -499,123 +423,175 @@ static int release(char *master_ifname, char *slave_ifname)
 	}
 
 	/* set to default mtu */
-	set_slave_mtu(slave_ifname, 1500);
+	set_mtu(slave_ifname, 1500);
 
 	return res;
 }
 
-static int get_if_settings(char *ifname, struct dev_data *dd)
-{
-	int res;
-
-	strncpy(dd->mtu.ifr_name, ifname, IFNAMSIZ);
-	res = ioctl(skfd, SIOCGIFMTU, &dd->mtu);
-	strncpy(dd->flags.ifr_name, ifname, IFNAMSIZ);
-	res |= ioctl(skfd, SIOCGIFFLAGS, &dd->flags);
-	strncpy(dd->hwaddr.ifr_name, ifname, IFNAMSIZ);
-	res |= ioctl(skfd, SIOCGIFHWADDR, &dd->hwaddr);
-
-	return res;
-}
-
-static int get_slave_flags(char *slave_ifname)
-{
-	strncpy(slave.flags.ifr_name, slave_ifname, IFNAMSIZ);
-	return ioctl(skfd, SIOCGIFFLAGS, &slave.flags);
-}
-
-static int set_master_hwaddr(char *master_ifname, struct sockaddr *hwaddr)
+static NOINLINE void get_drv_info(char *master_ifname)
 {
 	struct ifreq ifr;
+	struct ethtool_drvinfo info;
 
-	strncpy(ifr.ifr_name, master_ifname, IFNAMSIZ);
-	memcpy(&(ifr.ifr_hwaddr), hwaddr, sizeof(struct sockaddr));
-	return ioctl(skfd, SIOCSIFHWADDR, &ifr);
-}
-
-static int set_slave_hwaddr(char *slave_ifname, struct sockaddr *hwaddr)
-{
-	struct ifreq ifr;
-
-	strncpy(ifr.ifr_name, slave_ifname, IFNAMSIZ);
-	memcpy(&(ifr.ifr_hwaddr), hwaddr, sizeof(struct sockaddr));
-	return ioctl(skfd, SIOCSIFHWADDR, &ifr);
-}
-
-static int set_slave_mtu(char *slave_ifname, int mtu)
-{
-	struct ifreq ifr;
-
-	ifr.ifr_mtu = mtu;
-	strncpy(ifr.ifr_name, slave_ifname, IFNAMSIZ);
-	return ioctl(skfd, SIOCSIFMTU, &ifr);
-}
-
-static int set_if_flags(char *ifname, short flags)
-{
-	struct ifreq ifr;
-
-	ifr.ifr_flags = flags;
-	strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
-	return ioctl(skfd, SIOCSIFFLAGS, &ifr);
-}
-
-static int set_if_up(char *ifname, short flags)
-{
-	int res = set_if_flags(ifname, flags | IFF_UP);
-	if (res)
-		bb_perror_msg("%s: can't up", ifname);
-	return res;
-}
-
-static int set_if_down(char *ifname, short flags)
-{
-	int res = set_if_flags(ifname, flags & ~IFF_UP);
-	if (res)
-		bb_perror_msg("%s: can't down", ifname);
-	return res;
-}
-
-static int clear_if_addr(char *ifname)
-{
-	struct ifreq ifr;
-
-	strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
-	ifr.ifr_addr.sa_family = AF_INET;
-	memset(ifr.ifr_addr.sa_data, 0, sizeof(ifr.ifr_addr.sa_data));
-	return ioctl(skfd, SIOCSIFADDR, &ifr);
-}
-
-static int set_if_addr(char *master_ifname, char *slave_ifname)
-{
-	static const struct {
-		int g_ioctl;
-		int s_ioctl;
-	} ifra[] = {
-		{ SIOCGIFADDR,    SIOCSIFADDR    },
-		{ SIOCGIFDSTADDR, SIOCSIFDSTADDR },
-		{ SIOCGIFBRDADDR, SIOCSIFBRDADDR },
-		{ SIOCGIFNETMASK, SIOCSIFNETMASK },
-	};
-
-	struct ifreq ifr;
-	int res;
-	int i;
-
-	for (i = 0; i < ARRAY_SIZE(ifra); i++) {
-		strncpy(ifr.ifr_name, master_ifname, IFNAMSIZ);
-		res = ioctl(skfd, ifra[i].g_ioctl, &ifr);
-		if (res < 0) {
-			ifr.ifr_addr.sa_family = AF_INET;
-			memset(ifr.ifr_addr.sa_data, 0,
-			       sizeof(ifr.ifr_addr.sa_data));
-		}
-
-		strncpy(ifr.ifr_name, slave_ifname, IFNAMSIZ);
-		res = ioctl(skfd, ifra[i].s_ioctl, &ifr);
-		if (res < 0)
-			return res;
+	memset(&ifr, 0, sizeof(ifr));
+	ifr.ifr_data = (caddr_t)&info;
+	info.cmd = ETHTOOL_GDRVINFO;
+	/* both fields are 32 bytes long (long enough) */
+	strcpy(info.driver, "ifenslave");
+	strcpy(info.fw_version, utoa(BOND_ABI_VERSION));
+	if (set_ifrname_and_do_ioctl(SIOCETHTOOL, &ifr, master_ifname) < 0) {
+		if (errno == EOPNOTSUPP)
+			return;
+		bb_perror_msg_and_die("%s: SIOCETHTOOL error", master_ifname);
 	}
 
-	return 0;
+	abi_ver = bb_strtou(info.fw_version, NULL, 0);
+	if (errno)
+		bb_error_msg_and_die("%s: SIOCETHTOOL error", master_ifname);
+}
+
+int ifenslave_main(int argc, char **argv) MAIN_EXTERNALLY_VISIBLE;
+int ifenslave_main(int argc UNUSED_PARAM, char **argv)
+{
+	char *master_ifname, *slave_ifname;
+	int rv;
+	int res;
+	unsigned opt;
+	enum {
+		OPT_c = (1 << 0),
+		OPT_d = (1 << 1),
+		OPT_f = (1 << 2),
+	};
+#if ENABLE_GETOPT_LONG
+	static const char ifenslave_longopts[] ALIGN1 =
+		"change-active\0"  No_argument "c"
+		"detach\0"         No_argument "d"
+		"force\0"          No_argument "f"
+		/* "all-interfaces\0" No_argument "a" */
+		;
+
+	applet_long_options = ifenslave_longopts;
+#endif
+	INIT_G();
+
+	opt = getopt32(argv, "cdfa");
+	argv += optind;
+	if (opt & (opt-1)) /* Only one option can be given */
+		bb_show_usage();
+
+	master_ifname = *argv++;
+
+	/* No interface names - show all interfaces. */
+	if (!master_ifname) {
+		display_interfaces(NULL);
+		return EXIT_SUCCESS;
+	}
+
+	/* Open a basic socket */
+	xmove_fd(xsocket(AF_INET, SOCK_DGRAM, 0), skfd);
+
+	/* Exchange abi version with bonding module */
+	get_drv_info(master_ifname);
+
+	slave_ifname = *argv++;
+	if (!slave_ifname) {
+		if (opt & (OPT_d|OPT_c)) {
+			/* --change or --detach, and no slaves given -
+			 * show all interfaces. */
+			display_interfaces(slave_ifname /* == NULL */);
+			return 2; /* why 2? */
+		}
+		/* A single arg means show the
+		 * configuration for this interface
+		 */
+		display_interfaces(master_ifname);
+		return EXIT_SUCCESS;
+	}
+
+	if (get_if_settings(master_ifname, &master)) {
+		/* Probably a good reason not to go on */
+		bb_perror_msg_and_die("%s: can't get settings", master_ifname);
+	}
+
+	/* Check if master is indeed a master;
+	 * if not then fail any operation
+	 */
+	if (!(master.flags.ifr_flags & IFF_MASTER))
+		bb_error_msg_and_die("%s is not a master", master_ifname);
+
+	/* Check if master is up; if not then fail any operation */
+	if (!(master.flags.ifr_flags & IFF_UP))
+		bb_error_msg_and_die("%s is not up", master_ifname);
+
+#ifdef WHY_BOTHER
+	/* Neither -c[hange] nor -d[etach] -> it's "enslave" then;
+	 * and -f[orce] is not there too. Check that it's ethernet. */
+	if (!(opt & (OPT_d|OPT_c|OPT_f)) {
+		/* The family '1' is ARPHRD_ETHER for ethernet. */
+		if (master.hwaddr.ifr_hwaddr.sa_family != 1) {
+			bb_error_msg_and_die(
+				"%s is not ethernet-like (-f overrides)",
+				master_ifname);
+		}
+	}
+#endif
+
+	/* Accepts only one slave */
+	if (opt & OPT_c) {
+		/* Change active slave */
+		if (get_slave_flags(slave_ifname)) {
+			bb_perror_msg_and_die(
+				"%s: can't get flags", slave_ifname);
+		}
+		change_active(master_ifname, slave_ifname);
+		return EXIT_SUCCESS;
+	}
+
+	/* Accepts multiple slaves */
+	res = 0;
+	do {
+		if (opt & OPT_d) {
+			/* Detach a slave interface from the master */
+			rv = get_slave_flags(slave_ifname);
+			if (rv) {
+				/* Can't work with this slave, */
+				/* remember the error and skip it */
+				bb_perror_msg(
+					"skipping %s: can't get flags",
+					slave_ifname);
+				res = rv;
+				continue;
+			}
+			rv = release(master_ifname, slave_ifname);
+			if (rv) {
+				bb_perror_msg("can't release %s from %s",
+					slave_ifname, master_ifname);
+				res = rv;
+			}
+		} else {
+			/* Attach a slave interface to the master */
+			rv = get_if_settings(slave_ifname, &slave);
+			if (rv) {
+				/* Can't work with this slave, */
+				/* remember the error and skip it */
+				bb_perror_msg(
+					"skipping %s: can't get settings",
+					slave_ifname);
+				res = rv;
+				continue;
+			}
+			rv = enslave(master_ifname, slave_ifname);
+			if (rv) {
+				bb_perror_msg("can't enslave %s to %s",
+					slave_ifname, master_ifname);
+				res = rv;
+			}
+		}
+	} while ((slave_ifname = *argv++) != NULL);
+
+	if (ENABLE_FEATURE_CLEAN_UP) {
+		close(skfd);
+	}
+
+	return res;
 }
