@@ -23,7 +23,7 @@ struct dhcpOfferedAddr *leases;
 
 
 int udhcpd_main(int argc, char **argv) MAIN_EXTERNALLY_VISIBLE;
-int udhcpd_main(int argc ATTRIBUTE_UNUSED, char **argv)
+int udhcpd_main(int argc UNUSED_PARAM, char **argv)
 {
 	fd_set rfds;
 	struct timeval tv;
@@ -94,7 +94,7 @@ int udhcpd_main(int argc ATTRIBUTE_UNUSED, char **argv)
 	leases = xzalloc(server_config.max_leases * sizeof(*leases));
 	read_leases(server_config.lease_file);
 
-	if (read_interface(server_config.interface, &server_config.ifindex,
+	if (udhcp_read_interface(server_config.interface, &server_config.ifindex,
 			   &server_config.server, server_config.arp)) {
 		retval = 1;
 		goto ret;
@@ -107,7 +107,7 @@ int udhcpd_main(int argc ATTRIBUTE_UNUSED, char **argv)
 	while (1) { /* loop until universe collapses */
 
 		if (server_socket < 0) {
-			server_socket = listen_socket(/*INADDR_ANY,*/ SERVER_PORT,
+			server_socket = udhcp_listen_socket(/*INADDR_ANY,*/ SERVER_PORT,
 					server_config.interface);
 		}
 
@@ -145,7 +145,7 @@ int udhcpd_main(int argc ATTRIBUTE_UNUSED, char **argv)
 		default: continue;	/* signal or error (probably EINTR) */
 		}
 
-		bytes = udhcp_recv_packet(&packet, server_socket); /* this waits for a packet - idle */
+		bytes = udhcp_recv_kernel_packet(&packet, server_socket); /* this waits for a packet - idle */
 		if (bytes < 0) {
 			if (bytes == -1 && errno != EINTR) {
 				DEBUG("error on read, %s, reopening socket", strerror(errno));
@@ -180,7 +180,7 @@ int udhcpd_main(int argc ATTRIBUTE_UNUSED, char **argv)
 		case DHCPDISCOVER:
 			DEBUG("Received DISCOVER");
 
-			if (sendOffer(&packet) < 0) {
+			if (send_offer(&packet) < 0) {
 				bb_error_msg("send OFFER failed");
 			}
 			break;
@@ -200,20 +200,19 @@ int udhcpd_main(int argc ATTRIBUTE_UNUSED, char **argv)
 					if (server_id_align == server_config.server && requested
 					 && requested_align == lease->yiaddr
 					) {
-						sendACK(&packet, lease->yiaddr);
+						send_ACK(&packet, lease->yiaddr);
 					}
 				} else if (requested) {
 					/* INIT-REBOOT State */
 					if (lease->yiaddr == requested_align)
-						sendACK(&packet, lease->yiaddr);
+						send_ACK(&packet, lease->yiaddr);
 					else
-						sendNAK(&packet);
+						send_NAK(&packet);
 				} else if (lease->yiaddr == packet.ciaddr) {
 					/* RENEWING or REBINDING State */
-					sendACK(&packet, lease->yiaddr);
-				} else {
-					/* don't know what to do!!!! */
-					sendNAK(&packet);
+					send_ACK(&packet, lease->yiaddr);
+				} else { /* don't know what to do!!!! */
+					send_NAK(&packet);
 				}
 
 			/* what to do if we have no record of the client */
@@ -229,13 +228,13 @@ int udhcpd_main(int argc ATTRIBUTE_UNUSED, char **argv)
 						memset(lease->chaddr, 0, 16);
 					/* make some contention for this address */
 					} else
-						sendNAK(&packet);
+						send_NAK(&packet);
 				} else {
 					uint32_t r = ntohl(requested_align);
 					if (r < server_config.start_ip
 				         || r > server_config.end_ip
 					) {
-						sendNAK(&packet);
+						send_NAK(&packet);
 					}
 					/* else remain silent */
 				}
