@@ -6,7 +6,6 @@
  *
  */
 
-#include <getopt.h>	/* for struct option */
 #include "libbb.h"
 
 struct host_info {
@@ -55,9 +54,9 @@ enum {
 	STALLTIME = 5                   /* Seconds when xfer considered "stalled" */
 };
 
-static int getttywidth(void)
+static unsigned int getttywidth(void)
 {
-	int width;
+	unsigned width;
 	get_terminal_width_height(0, &width, NULL);
 	return width;
 }
@@ -189,7 +188,7 @@ static void progressmeter(int flag)
  */
 #else /* FEATURE_WGET_STATUSBAR */
 
-static ALWAYS_INLINE void progressmeter(int flag ATTRIBUTE_UNUSED) { }
+static ALWAYS_INLINE void progressmeter(int flag UNUSED_PARAM) { }
 
 #endif
 
@@ -387,7 +386,7 @@ static char *gethdr(char *buf, size_t bufsiz, FILE *fp /*, int *istrunc*/)
 
 
 int wget_main(int argc, char **argv) MAIN_EXTERNALLY_VISIBLE;
-int wget_main(int argc ATTRIBUTE_UNUSED, char **argv)
+int wget_main(int argc UNUSED_PARAM, char **argv)
 {
 	char buf[512];
 	struct host_info server, target;
@@ -418,15 +417,17 @@ int wget_main(int argc ATTRIBUTE_UNUSED, char **argv)
 		KEY_content_length = 1, KEY_transfer_encoding, KEY_chunked, KEY_location
 	};
 	enum {
-		WGET_OPT_CONTINUE   = 0x1,
-		WGET_OPT_SPIDER	    = 0x2,
-		WGET_OPT_QUIET      = 0x4,
-		WGET_OPT_OUTNAME    = 0x8,
-		WGET_OPT_PREFIX     = 0x10,
-		WGET_OPT_PROXY      = 0x20,
-		WGET_OPT_USER_AGENT = 0x40,
-		WGET_OPT_PASSIVE    = 0x80,
-		WGET_OPT_HEADER     = 0x100,
+		WGET_OPT_CONTINUE   = (1 << 0),
+		WGET_OPT_SPIDER	    = (1 << 1),
+		WGET_OPT_QUIET      = (1 << 2),
+		WGET_OPT_OUTNAME    = (1 << 3),
+		WGET_OPT_PREFIX     = (1 << 4),
+		WGET_OPT_PROXY      = (1 << 5),
+		WGET_OPT_USER_AGENT = (1 << 6),
+		WGET_OPT_RETRIES    = (1 << 7),
+		WGET_OPT_NETWORK_READ_TIMEOUT = (1 << 8),
+		WGET_OPT_PASSIVE    = (1 << 9),
+		WGET_OPT_HEADER     = (1 << 10),
 	};
 #if ENABLE_FEATURE_WGET_LONG_OPTIONS
 	static const char wget_longopts[] ALIGN1 =
@@ -438,6 +439,10 @@ int wget_main(int argc ATTRIBUTE_UNUSED, char **argv)
 		"directory-prefix\0" Required_argument "P"
 		"proxy\0"            Required_argument "Y"
 		"user-agent\0"       Required_argument "U"
+		/* Ignored: */
+		// "tries\0"            Required_argument "t"
+		// "timeout\0"          Required_argument "T"
+		/* Ignored (we always use PASV): */
 		"passive-ftp\0"      No_argument       "\xff"
 		"header\0"           Required_argument "\xfe"
 		;
@@ -450,9 +455,11 @@ int wget_main(int argc ATTRIBUTE_UNUSED, char **argv)
 #endif
 	/* server.allocated = target.allocated = NULL; */
 	opt_complementary = "-1" USE_FEATURE_WGET_LONG_OPTIONS(":\xfe::");
-	opt = getopt32(argv, "csqO:P:Y:U:",
+	opt = getopt32(argv, "csqO:P:Y:U:" /*ignored:*/ "t:T:",
 				&fname_out, &dir_prefix,
-				&proxy_flag, &user_agent
+				&proxy_flag, &user_agent,
+				NULL, /* -t RETRIES */
+				NULL /* -T NETWORK_READ_TIMEOUT */
 				USE_FEATURE_WGET_LONG_OPTIONS(, &headers_llist)
 				);
 	if (strcmp(proxy_flag, "off") == 0) {
@@ -470,8 +477,7 @@ int wget_main(int argc ATTRIBUTE_UNUSED, char **argv)
 		}
 		extra_headers = cp = xmalloc(size);
 		while (headers_llist) {
-			cp += sprintf(cp, "%s\r\n", headers_llist->data);
-			headers_llist = headers_llist->link;
+			cp += sprintf(cp, "%s\r\n", (char*)llist_pop(&headers_llist));
 		}
 	}
 #endif
@@ -600,6 +606,31 @@ int wget_main(int argc ATTRIBUTE_UNUSED, char **argv)
 					/* eat all remaining headers */;
 				goto read_response;
 			case 200:
+/*
+Response 204 doesn't say "null file", it says "metadata
+has changed but data didn't":
+
+"10.2.5 204 No Content
+The server has fulfilled the request but does not need to return
+an entity-body, and might want to return updated metainformation.
+The response MAY include new or updated metainformation in the form
+of entity-headers, which if present SHOULD be associated with
+the requested variant.
+
+If the client is a user agent, it SHOULD NOT change its document
+view from that which caused the request to be sent. This response
+is primarily intended to allow input for actions to take place
+without causing a change to the user agent's active document view,
+although any new or updated metainformation SHOULD be applied
+to the document currently in the user agent's active view.
+
+The 204 response MUST NOT include a message-body, and thus
+is always terminated by the first empty line after the header fields."
+
+However, in real world it was observed that some web servers
+(e.g. Boa/0.94.14rc21) simply use code 204 when file size is zero.
+*/
+			case 204:
 				break;
 			case 300:	/* redirection */
 			case 301:
